

Imports System.ComponentModel
Imports System.Diagnostics


' Meta messages are MIDI messages that are stored in MIDI files. 
' These messages are not sent or received via MIDI but are read and 
' interpretted from MIDI files. 
' They provide information that describes a MIDI file's properties. 
' For example, tempo changes are implemented using meta messages.

' ---------------------------------------------------------------------------------- MetaMessage types
Public Enum MetaType
    SequenceNumber
    Text
    Copyright
    TrackName
    InstrumentName
    Lyric
    Marker
    CuePoint
    ProgramName
    DeviceName
    EndOfTrack = &H2F
    Tempo = &H51
    SmpteOffset = &H54
    TimeSignature = &H58
    KeySignature
    ProprietaryEvent = &H7F
End Enum


<ImmutableObject(True)> _
Public NotInheritable Class MetaMessage
	Implements IMidiMessage

    ' ---------------------------------------------------------- shift data bytes when calculating the hash code
    Private Const Shift As Integer = 7
    ' ---------------------------------------------------------- length in bytes for tempo meta message data
    Public Const TempoLength As Integer = 3
    ' ---------------------------------------------------------- length in bytes for SMPTE offset meta message data
    Public Const SmpteOffsetLength As Integer = 5
    ' ---------------------------------------------------------- length in bytes for time signature meta message data
    Public Const TimeSigLength As Integer = 4
    ' ---------------------------------------------------------- Length in bytes for key signature meta message data
    Public Const KeySigLength As Integer = 2


    ' ---------------------------------------------------------- End of track meta message.
    Public Shared ReadOnly EndOfTrackMessage As New MetaMessage(MetaType.EndOfTrack, New Byte(-1) {})

    ' ---------------------------------------------------------- The meta message type.
    Private type As MetaType
    ' ---------------------------------------------------------- The meta message data.
    Private data As Byte()
    ' ---------------------------------------------------------- The hash code value.
    Private hashCode As Integer


    ' Each MetaMessage has type and length properties. 
    ' For certain types, the length of the message data must be a specific value.
    ' For example, tempo messages must have a data length of exactly three. 
    ' Some MetaMessage types can have any data length. 
    ' Text messages are an example of a MetaMessage that can have a variable data length.
    ' When a MetaMessage is created, the length of the data is checked
    ' to make sure that it is valid for the specified type. 
    ' If it is not, an exception is thrown. 

    Public Sub New(ByVal type As MetaType, ByVal data As Byte())
        If data Is Nothing Then
            Throw New ArgumentNullException("data")
        ElseIf Not ValidateDataLength(type, data.Length) Then
            Throw New ArgumentException("Length of data not valid for meta message type.")
        End If
        '
        Me.type = type
        ' --------------------------------------------------------- Create storage for meta message data.
        Me.data = New Byte(data.Length - 1) {}
        ' --------------------------------------------------------- Copy data into storage.
        data.CopyTo(Me.data, 0)
        '
        CalculateHashCode()
    End Sub

    ' ------------------------------------------------------------- gets a copy of the data bytes for this meta message.
    Public Function GetBytes() As Byte() Implements IMidiMessage.GetBytes
        Return DirectCast(data.Clone(), Byte())
    End Function

    ' ------------------------------------------------------------- get a value for the current MetaMessage 
    '                                                               suitable for use in hashing algorithms.
    Public Overloads Overrides Function GetHashCode() As Integer
        Return hashCode
    End Function

    ' -------------------------------------------------------------- determines if MetaMessage instances are equal
    Public Overloads Overrides Function Equals(ByVal obj As Object) As Boolean
        If Not (TypeOf obj Is MetaMessage) Then
            Return False
        End If

        Dim equal As Boolean = True
        Dim message As MetaMessage = DirectCast(obj, MetaMessage)

        ' --------------------------------------------------------- if the types do not match
        If MetaType <> message.MetaType Then
            equal = False
        End If

        ' --------------------------------------------------------- if message lengths are not equal
        If equal AndAlso Length <> message.Length Then
            equal = False
        End If

        ' --------------------------------------------------------- check if data are equal
        Dim i As Integer = 0
        While i < Length AndAlso equal
            If Me(i) <> message(i) Then
                equal = False
            End If
            i += 1
        End While

        Return equal
    End Function


    Private Sub CalculateHashCode()
        ' -------------------------------------------- This algorithm may need work
        hashCode = MetaType
        For i As Integer = 0 To data.Length - 1 Step 3
            hashCode = hashCode Xor data(i)
        Next
        For i As Integer = 1 To data.Length - 1 Step 3
            hashCode = hashCode Xor CInt(data(i)) << Shift
        Next
        For i As Integer = 2 To data.Length - 1 Step 3
            hashCode = hashCode Xor CInt(data(i)) << Shift * 2
        Next
    End Sub

 
    ' ----------------------------------------------- test if the data length is valid for this type of MetaMessage
    Private Function ValidateDataLength(ByVal type As MetaType, ByVal length As Integer) As Boolean
        Debug.Assert(length >= 0)
        '
        Dim result As Boolean = True
        ' ------------------------------------------------------------ determine which type of meta message this is
        '                                                              and check if data length value is valid
        Select Case type
            Case MetaType.SequenceNumber
                If length <> 0 OrElse length <> 2 Then result = False

            Case MetaType.EndOfTrack
                If length <> 0 Then result = False

            Case MetaType.Tempo
                If length <> TempoLength Then result = False

            Case MetaType.SmpteOffset
                If length <> SmpteOffsetLength Then result = False

            Case MetaType.TimeSignature
                If length <> TimeSigLength Then result = False

            Case MetaType.KeySignature
                If length <> KeySigLength Then result = False

            Case Else
                result = True

        End Select
        '
        Return result
    End Function


    Default Public ReadOnly Property Item(ByVal index As Integer) As Byte
        Get
            If index < 0 OrElse index >= Length Then
                Throw New ArgumentOutOfRangeException("index", index, "Index into MetaMessage out of range.")
            End If

            Return data(index)
        End Get
    End Property

    ' ------------------------------------------------------------- gets the length of the meta message
    Public ReadOnly Property Length() As Integer
        Get
            Return data.Length
        End Get
    End Property

    ' -------------------------------------------------------------- gets the type of meta message
    Public ReadOnly Property MetaType() As MetaType
        Get
            Return type
        End Get
    End Property


    Public ReadOnly Property Status() As Integer Implements IMidiMessage.Status
        Get
            ' All meta messages have the same status value (0xFF).
            Return &HFF
        End Get
    End Property

    Public ReadOnly Property MessageType() As MessageType Implements IMidiMessage.MessageType
        Get
            Return MessageType.Meta
        End Get
    End Property


End Class
