

Imports System.Collections
Imports System.Collections.Generic
Imports System.ComponentModel
Imports System.Diagnostics
Imports System.IO

' ------------------------------------------------------------------- a collection of Tracks
Public NotInheritable Class Sequence
    Implements IComponent

    Implements ICollection(Of Track)


    ' --------------------------------------------------------------- the collection of Tracks for the Sequence
    Private tracks As New List(Of Track)()
    ' --------------------------------------------------------------- the Sequence's MIDI file properties
    Private properties As New MidiFileProperties()
    Private loadWorker As New BackgroundWorker()
    Private saveWorker As New BackgroundWorker()
    Private m_site As ISite = Nothing
    Private m_disposed As Boolean = False

    Public Event LoadCompleted As EventHandler(Of AsyncCompletedEventArgs)
    Public Event LoadProgressChanged As ProgressChangedEventHandler
    Public Event SaveCompleted As EventHandler(Of AsyncCompletedEventArgs)
    Public Event SaveProgressChanged As ProgressChangedEventHandler

    Public Sub New()
        InitializeBackgroundWorkers()
    End Sub

    Public Sub New(ByVal division As Integer)
        properties.Division = division
        properties.Format = 1
        InitializeBackgroundWorkers()
    End Sub

    Public Sub New(ByVal fileName As String)
        InitializeBackgroundWorkers()
        Load(fileName)
    End Sub


    Private Sub InitializeBackgroundWorkers()
        AddHandler loadWorker.DoWork, New DoWorkEventHandler(AddressOf LoadDoWork)
        AddHandler loadWorker.ProgressChanged, New ProgressChangedEventHandler(AddressOf OnLoadProgressChanged)
        AddHandler loadWorker.RunWorkerCompleted, New RunWorkerCompletedEventHandler(AddressOf OnLoadCompleted)
        loadWorker.WorkerReportsProgress = True

        AddHandler saveWorker.DoWork, New DoWorkEventHandler(AddressOf SaveDoWork)
        AddHandler saveWorker.ProgressChanged, New ProgressChangedEventHandler(AddressOf OnSaveProgressChanged)
        AddHandler saveWorker.RunWorkerCompleted, New RunWorkerCompletedEventHandler(AddressOf OnSaveCompleted)
        saveWorker.WorkerReportsProgress = True
    End Sub


    ' ------------------------------------------------------------------------------ "Methods"


    ' ----------------------------------------------------- Loads a MIDI file into the Sequence.
    '
    Public Sub Load(ByVal fileName As String)
        If m_disposed Then
            Throw New ObjectDisposedException("Sequence")
        ElseIf IsBusy Then
            Throw New InvalidOperationException()
        ElseIf fileName Is Nothing Then
            Throw New ArgumentNullException("fileName")
        End If

        Dim stream As New FileStream(fileName, FileMode.Open, FileAccess.Read, FileShare.Read)

        Using stream
            Dim newProperties As New MidiFileProperties()
            Dim reader As New TrackReader()
            Dim newTracks As New List(Of Track)()

            newProperties.Read(stream)

            For i As Integer = 0 To newProperties.TrackCount - 1
                reader.Read(stream)
                newTracks.Add(reader.Track)
            Next

            properties = newProperties
            tracks = newTracks
        End Using

        Debug.Assert(Count = properties.TrackCount)
    End Sub

    Public Sub LoadAsync(ByVal fileName As String)
        If m_disposed Then
            Throw New ObjectDisposedException("Sequence")
        ElseIf IsBusy Then
            Throw New InvalidOperationException()
        ElseIf fileName Is Nothing Then
            Throw New ArgumentNullException("fileName")
        End If
        loadWorker.RunWorkerAsync(fileName)
    End Sub

    Public Sub LoadAsyncCancel()
        If m_disposed Then
            Throw New ObjectDisposedException("Sequence")
        End If
        loadWorker.CancelAsync()
    End Sub


    ' ------------------------------------------------------------------- saves the Sequence as a MIDI file
    '
    Public Sub Save(ByVal fileName As String)
        If m_disposed Then
            Throw New ObjectDisposedException("Sequence")
        ElseIf fileName Is Nothing Then
            Throw New ArgumentNullException("fileName")
        End If

        Dim stream As New FileStream(fileName, FileMode.Create, FileAccess.Write, FileShare.None)

        Using stream
            properties.Write(stream)

            Dim writer As New TrackWriter()

            For Each trk As Track In tracks
                writer.Track = trk
                writer.Write(stream)
            Next
        End Using
    End Sub

    Public Sub SaveAsync(ByVal fileName As String)
        If m_disposed Then
            Throw New ObjectDisposedException("Sequence")
        ElseIf IsBusy Then
            Throw New InvalidOperationException()
        ElseIf fileName Is Nothing Then
            Throw New ArgumentNullException("fileName")
        End If

        saveWorker.RunWorkerAsync(fileName)
    End Sub

    Public Sub SaveAsyncCancel()
        If m_disposed Then
            Throw New ObjectDisposedException("Sequence")
        End If

        saveWorker.CancelAsync()
    End Sub


    ' ------------------------------------------------------------------ gets the length in ticks of the Sequence
    '
    Public Function GetLength() As Integer
        If m_disposed Then
            Throw New ObjectDisposedException("Sequence")
        End If

        Dim length As Integer = 0

        For Each t As Track In Me
            If t.Length > length Then
                length = t.Length
            End If
        Next

        Return length
    End Function

    Private Sub OnLoadCompleted(ByVal sender As Object, ByVal e As RunWorkerCompletedEventArgs)
        RaiseEvent LoadCompleted(Me, New AsyncCompletedEventArgs(e.[Error], e.Cancelled, Nothing))
    End Sub

    Private Sub OnLoadProgressChanged(ByVal sender As Object, ByVal e As ProgressChangedEventArgs)
        RaiseEvent LoadProgressChanged(Me, e)
    End Sub

    Private Sub LoadDoWork(ByVal sender As Object, ByVal e As DoWorkEventArgs)
        Dim fileName As String = DirectCast(e.Argument, String)

        Dim stream As New FileStream(fileName, FileMode.Open, FileAccess.Read, FileShare.Read)

        Using stream
            Dim newProperties As New MidiFileProperties()
            Dim reader As New TrackReader()
            Dim newTracks As New List(Of Track)()

            newProperties.Read(stream)

            Dim percentage As Single

            Dim i As Integer = 0
            While i < newProperties.TrackCount AndAlso Not loadWorker.CancellationPending
                reader.Read(stream)
                newTracks.Add(reader.Track)
                '
                percentage = (i + 1.0F) / newProperties.TrackCount
                loadWorker.ReportProgress(CInt(100 * percentage))
                i += 1
            End While

            If loadWorker.CancellationPending Then
                e.Cancel = True
            Else
                properties = newProperties
                tracks = newTracks
            End If
        End Using
    End Sub

    Private Sub OnSaveCompleted(ByVal sender As Object, ByVal e As RunWorkerCompletedEventArgs)
        RaiseEvent SaveCompleted(Me, New AsyncCompletedEventArgs(e.[Error], e.Cancelled, Nothing))
    End Sub

    Private Sub OnSaveProgressChanged(ByVal sender As Object, ByVal e As ProgressChangedEventArgs)
        RaiseEvent SaveProgressChanged(Me, e)
    End Sub

    Private Sub SaveDoWork(ByVal sender As Object, ByVal e As DoWorkEventArgs)
        Dim fileName As String = DirectCast(e.Argument, String)

        Dim stream As New FileStream(fileName, FileMode.Create, FileAccess.Write, FileShare.None)

        Using stream
            properties.Write(stream)

            Dim writer As New TrackWriter()

            Dim percentage As Single

            Dim i As Integer = 0
            While i < tracks.Count AndAlso Not saveWorker.CancellationPending
                writer.Track = tracks(i)
                writer.Write(stream)

                percentage = (i + 1.0F) / properties.TrackCount

                saveWorker.ReportProgress(CInt(100 * percentage))
                i += 1
            End While

            If saveWorker.CancellationPending Then
                e.Cancel = True
            End If
        End Using
    End Sub


    ' -------------------------------------------------------------------------- "Properties"


    ' --------------------------------------------------------------------- gets the Track at the specified index
    '
    Default Public ReadOnly Property Item(ByVal index As Integer) As Track
        Get
            If m_disposed Then
                Throw New ObjectDisposedException("Sequence")
            ElseIf index < 0 OrElse index >= Count Then
                Throw New ArgumentOutOfRangeException("index", index, "Sequence index out of range.")
            End If

            Return tracks(index)
        End Get
    End Property

    ' -------------------------------------------------------------------- gets the Sequence's division value
    '
    Public ReadOnly Property Division() As Integer
        Get
            If m_disposed Then
                Throw New ObjectDisposedException("Sequence")
            End If

            Return properties.Division
        End Get
    End Property


    ' -------------------------------------------------------------------- gets or sets the Sequence's format value
    '
    Public Property Format() As Integer
        Get
            If m_disposed Then
                Throw New ObjectDisposedException("Sequence")
            End If

            Return properties.Format
        End Get
        Set(ByVal value As Integer)
            If m_disposed Then
                Throw New ObjectDisposedException("Sequence")
            ElseIf IsBusy Then
                Throw New InvalidOperationException()
            End If

            properties.Format = value
        End Set
    End Property


    ' ---------------------------------------------------------------- gets the Sequence's type
    '
    Public ReadOnly Property SequenceType() As SequenceType
        Get
            If m_disposed Then
                Throw New ObjectDisposedException("Sequence")
            End If

            Return properties.SequenceType
        End Get
    End Property

    Public ReadOnly Property IsBusy() As Boolean
        Get
            Return loadWorker.IsBusy OrElse saveWorker.IsBusy
        End Get
    End Property



#Region "ICollection<Track> Members"

    Public Sub Add(ByVal item As Track) Implements System.Collections.Generic.ICollection(Of Track).Add
        If m_disposed Then
            Throw New ObjectDisposedException("Sequence")
        ElseIf item Is Nothing Then
            Throw New ArgumentNullException("item")
        End If
        tracks.Add(item)
        properties.TrackCount = tracks.Count
    End Sub

    Public Sub Clear() Implements ICollection(Of Track).Clear
        If m_disposed Then
            Throw New ObjectDisposedException("Sequence")
        End If
        tracks.Clear()
        properties.TrackCount = tracks.Count
    End Sub

    Public Function Contains(ByVal item As Track) As Boolean Implements System.Collections.Generic.ICollection(Of Track).Contains
        If m_disposed Then
            Throw New ObjectDisposedException("Sequence")
        End If
        Return tracks.Contains(item)
    End Function

    Public Sub CopyTo(ByVal array As Track(), ByVal arrayIndex As Integer) Implements System.Collections.Generic.ICollection(Of Track).CopyTo
        If m_disposed Then
            Throw New ObjectDisposedException("Sequence")
        End If
        tracks.CopyTo(array, arrayIndex)
    End Sub

    Public ReadOnly Property Count() As Integer Implements ICollection(Of Track).Count
        Get
            If m_disposed Then
                Throw New ObjectDisposedException("Sequence")
            End If
            Return tracks.Count
        End Get
    End Property

    Public ReadOnly Property IsReadOnly() As Boolean Implements ICollection(Of Track).IsReadOnly
        Get
            If m_disposed Then
                Throw New ObjectDisposedException("Sequence")
            End If
            Return False
        End Get
    End Property

    Public Function Remove(ByVal item As Track) As Boolean Implements System.Collections.Generic.ICollection(Of Track).Remove
        If m_disposed Then
            Throw New ObjectDisposedException("Sequence")
        End If

        Dim result As Boolean = tracks.Remove(item)

        If result Then
            properties.TrackCount = tracks.Count
        End If

        Return result
    End Function

#End Region

#Region "IEnumerable<Track> Members"

    Public Function GetEnumerator() As IEnumerator(Of Track) Implements IEnumerable(Of Track).GetEnumerator
        If m_disposed Then
            Throw New ObjectDisposedException("Sequence")
        End If

        Return tracks.GetEnumerator()
    End Function

#End Region

#Region "IEnumerable Members"

    Private Function IEnumerable_GetEnumerator() As IEnumerator Implements IEnumerable.GetEnumerator
        If m_disposed Then
            Throw New ObjectDisposedException("Sequence")
        End If

        Return tracks.GetEnumerator()
    End Function

#End Region

#Region "IComponent Members"

    Public Event Disposed(ByVal sender As Object, ByVal e As System.EventArgs) Implements System.ComponentModel.IComponent.Disposed

    Public Property Site() As ISite Implements IComponent.Site
        Get
            Return m_site
        End Get
        Set(ByVal value As ISite)
            m_site = value
        End Set
    End Property

#End Region

#Region "IDisposable Members"

    Public Sub Dispose() Implements IDisposable.Dispose
        If m_disposed Then
            Return
        End If

        loadWorker.Dispose()
        saveWorker.Dispose()

        m_disposed = True

        RaiseEvent Disposed(Me, EventArgs.Empty)
    End Sub

#End Region


End Class
