

Imports System.Diagnostics

' ------------------------------------------------------ a collection of MidiEvents and a MIDI track within a Sequence
Public NotInheritable Partial Class Track
	
    ' -------------------------------------------------- the number of MidiEvents in the Track 
    '                                                    at least 1 ( Track always has an end of track message )
    Private m_count As Integer = 1
    ' -------------------------------------------------- the number of ticks to offset the end of track message
    Private m_endOfTrackOffset As Integer = 0
    ' -------------------------------------------------- the first MidiEvent in the Track
    Private head As MidiEvent = Nothing
    ' -------------------------------------------------- the last MidiEvent in the Track, 
    '                                                     ( not including the end of track message )
    Private tail As MidiEvent = Nothing
    ' -------------------------------------------------- the end of track MIDI event
    Private endOfTrackMidiEvent As MidiEvent


    Public Sub New()
        endOfTrackMidiEvent = New MidiEvent(Me, Length, MetaMessage.EndOfTrackMessage)
    End Sub


    ' -------------------------------------------------- Inserts IMidiMessage at specified position in absolute ticks
    Public Sub Insert(ByVal position As Integer, ByVal message As IMidiMessage)
        If position < 0 Then
            Throw New ArgumentOutOfRangeException("position", position, "IMidiMessage position out of range.")
        ElseIf message Is Nothing Then
            Throw New ArgumentNullException("message")
        End If

        Dim newMidiEvent As New MidiEvent(Me, position, message)

        If head Is Nothing Then
            head = newMidiEvent
            tail = newMidiEvent

        ElseIf position >= tail.AbsoluteTicks Then
            newMidiEvent.Previous = tail
            tail.[Next] = newMidiEvent
            tail = newMidiEvent
            endOfTrackMidiEvent.SetAbsoluteTicks(Length)
            endOfTrackMidiEvent.Previous = tail

        Else
            Dim current As MidiEvent = head
            While current.AbsoluteTicks < position
                current = current.[Next]
            End While
            newMidiEvent.[Next] = current
            newMidiEvent.Previous = current.Previous
            If current.Previous IsNot Nothing Then
                current.Previous.[Next] = newMidiEvent
            Else
                head = newMidiEvent
            End If
            current.Previous = newMidiEvent

        End If
        m_count += 1
    End Sub


    ' --------------------------------------- clears all of the MidiEvents, with the exception of the end of track
    Public Sub Clear()
        head = Nothing
        tail = Nothing
        m_count = 1
        AssertValid()
    End Sub


    ' --------------------------------------- merges the specified Track with the current Track
    Public Sub Merge(ByVal trk As Track)
        If trk Is Nothing Then
            Throw New ArgumentNullException("trk")
        End If

        If trk Is Me Then
            Return
        ElseIf trk.Count = 1 Then
            Return
        End If

#If Debug Then
        Dim oldCount As Integer = m_count
#End If

        m_count += trk.Count - 1

        Dim a As MidiEvent = head
        Dim b As MidiEvent = trk.head
        Dim current As MidiEvent = Nothing

        Debug.Assert(b IsNot Nothing)

        If a IsNot Nothing AndAlso a.AbsoluteTicks <= b.AbsoluteTicks Then
            current = New MidiEvent(Me, a.AbsoluteTicks, a.MidiMessage)
            a = a.[Next]
        Else
            current = New MidiEvent(Me, b.AbsoluteTicks, b.MidiMessage)
            b = b.[Next]
        End If

        head = current

        While a IsNot Nothing AndAlso b IsNot Nothing
            While a IsNot Nothing AndAlso a.AbsoluteTicks <= b.AbsoluteTicks
                current.[Next] = New MidiEvent(Me, a.AbsoluteTicks, a.MidiMessage)
                current.[Next].Previous = current
                current = current.[Next]
                a = a.[Next]
            End While

            If a IsNot Nothing Then
                While b IsNot Nothing AndAlso b.AbsoluteTicks <= a.AbsoluteTicks
                    current.[Next] = New MidiEvent(Me, b.AbsoluteTicks, b.MidiMessage)
                    current.[Next].Previous = current
                    current = current.[Next]
                    b = b.[Next]
                End While
            End If
        End While

        While a IsNot Nothing
            current.[Next] = New MidiEvent(Me, a.AbsoluteTicks, a.MidiMessage)
            current.[Next].Previous = current
            current = current.[Next]
            a = a.[Next]
        End While

        While b IsNot Nothing
            current.[Next] = New MidiEvent(Me, b.AbsoluteTicks, b.MidiMessage)
            current.[Next].Previous = current
            current = current.[Next]
            b = b.[Next]
        End While

        tail = current

        endOfTrackMidiEvent.SetAbsoluteTicks(Length)
        endOfTrackMidiEvent.Previous = tail

#If Debug Then
		Debug.Assert(m_count = oldCount + trk.Count - 1)
#End If

        AssertValid()
    End Sub


    ' -------------------------------------------------------- removes the MidiEvent at the specified index
    Public Sub RemoveAt(ByVal index As Integer)
        If index < 0 Then
            Throw New ArgumentOutOfRangeException("index", index, "Track index out of range.")
        ElseIf index = Count - 1 Then
            Throw New ArgumentException("Cannot remove the end of track event.", "index")
        End If

        Dim current As MidiEvent = GetMidiEvent(index)

        If current.Previous IsNot Nothing Then
            current.Previous.[Next] = current.[Next]
        Else
            Debug.Assert(current Is head)

            head = head.[Next]
        End If

        If current.[Next] IsNot Nothing Then
            current.[Next].Previous = current.Previous
        Else
            Debug.Assert(current Is tail)

            tail = tail.Previous

            endOfTrackMidiEvent.SetAbsoluteTicks(Length)
            endOfTrackMidiEvent.Previous = tail
        End If

        current.[Next] = Nothing
        current.Previous = Nothing

        m_count -= 1

        AssertValid()
    End Sub


    ' --------------------------------------------------------------- gets the MidiEvent at the specified index
    Public Function GetMidiEvent(ByVal index As Integer) As MidiEvent
        If index < 0 OrElse index >= Count Then
            Throw New ArgumentOutOfRangeException("index", index, "Track index out of range.")
        End If

        Dim result As MidiEvent

        If index = Count - 1 Then
            result = endOfTrackMidiEvent
        Else
            If index < Count \ 2 Then
                result = head

                For i As Integer = 0 To index - 1
                    result = result.[Next]
                Next
            Else
                result = tail

                For i As Integer = Count - 2 To index + 1 Step -1
                    result = result.Previous
                Next
            End If
        End If


#If Debug Then
		If index = Count - 1 Then
			Debug.Assert(result.AbsoluteTicks = Length)
			Debug.Assert(result.MidiMessage Is MetaMessage.EndOfTrackMessage)
		Else
			Dim t As MidiEvent = head
			For i As Integer = 0 To index - 1
				t = t.[Next]
			Next
			Debug.Assert(t Is result)
		End If
#End If

        Return result
    End Function


    Public Sub Move(ByVal e As MidiEvent, ByVal newPosition As Integer)
        If e.Owner IsNot Me Then
            Throw New ArgumentException("MidiEvent does not belong to this Track.")
        ElseIf newPosition < 0 Then
            Throw New ArgumentOutOfRangeException("newPosition")
        ElseIf e Is endOfTrackMidiEvent Then
            Throw New InvalidOperationException("Cannot move end of track message. Use the EndOfTrackOffset property instead.")
        End If

        Dim previous As MidiEvent = e.Previous
        Dim [next] As MidiEvent = e.[Next]

        If e.Previous IsNot Nothing AndAlso e.Previous.AbsoluteTicks > newPosition Then
            e.Previous.[Next] = e.[Next]

            If e.[Next] IsNot Nothing Then
                e.[Next].Previous = e.Previous
            End If

            While previous IsNot Nothing AndAlso previous.AbsoluteTicks > newPosition
                [next] = previous
                previous = previous.Previous
            End While
        ElseIf e.[Next] IsNot Nothing AndAlso e.[Next].AbsoluteTicks < newPosition Then
            e.[Next].Previous = e.Previous

            If e.Previous IsNot Nothing Then
                e.Previous.[Next] = e.[Next]
            End If

            While [next] IsNot Nothing AndAlso [next].AbsoluteTicks < newPosition
                previous = [next]
                [next] = [next].[Next]
            End While
        End If

        If previous IsNot Nothing Then
            previous.[Next] = e
        End If

        If [next] IsNot Nothing Then
            [next].Previous = e
        End If

        e.Previous = previous
        e.[Next] = [next]
        e.SetAbsoluteTicks(newPosition)

        If newPosition < head.AbsoluteTicks Then
            head = e
        End If

        If newPosition > tail.AbsoluteTicks Then
            tail = e
        End If

        endOfTrackMidiEvent.SetAbsoluteTicks(Length)
        endOfTrackMidiEvent.Previous = tail

        AssertValid()
    End Sub


    <Conditional("DEBUG")> _
    Private Sub AssertValid()
        Dim c As Integer = 1
        Dim current As MidiEvent = head
        Dim ticks As Integer = 1

        While current IsNot Nothing
            ticks += current.DeltaTicks

            If current.Previous IsNot Nothing Then
                Debug.Assert(current.AbsoluteTicks >= current.Previous.AbsoluteTicks)
                Debug.Assert(current.DeltaTicks = current.AbsoluteTicks - current.Previous.AbsoluteTicks)
            End If

            If current.[Next] Is Nothing Then
                Debug.Assert(tail Is current)
            End If

            current = current.[Next]

            c += 1
        End While

        ticks += EndOfTrackOffset

        Debug.Assert(ticks = Length, "Length mismatch")
        Debug.Assert(c = Count, "Count mismatch")
    End Sub



    ' --------------------------------------------------------------- get the number of MidiEvents in the Track
    Public ReadOnly Property Count() As Integer
        Get
            Return m_count
        End Get
    End Property

    ' --------------------------------------------------------------- get the length of the Track in ticks
    Public ReadOnly Property Length() As Integer
        Get
            Dim _length As Integer = EndOfTrackOffset
            If tail IsNot Nothing Then
                _length += tail.AbsoluteTicks
            End If
            Return _length + 1
        End Get
    End Property

    ' ---------------------------------------------------- get and set the end of track meta message position offset
    Public Property EndOfTrackOffset() As Integer
        Get
            Return m_endOfTrackOffset
        End Get
        Set(ByVal value As Integer)
            If value < 0 Then
                Throw New ArgumentOutOfRangeException("EndOfTrackOffset", value, "End of track offset out of range.")
            End If
            m_endOfTrackOffset = value
            endOfTrackMidiEvent.SetAbsoluteTicks(Length)
        End Set
    End Property

    ' ----------------------------------------------------- gets an object used to synchronize access to the Track
    Public ReadOnly Property SyncRoot() As Object
        Get
            Return Me
        End Get
    End Property

End Class
