

Imports System.Diagnostics

' -------------------------------------------------------------------------- represents a stereo chorus
Public Class Chorus
	Inherits EffectComponent
	Implements IProgrammable

    Public Enum ParameterId
        Delay
        Depth
        Feedback
        Frequency
        Phase
        Waveform
        WetDryMix
    End Enum

    Private Enum Phase
        Negative180
        Negative90
        Zero
        Positive90
        Positive180
    End Enum

    Private Const DelayLineMask As Integer = 4095
    Private Const DelayLineLength As Integer = DelayLineMask + 1
    Private Const DepthScaler As Single = 1
    Private Const FrequencyScaler As Integer = 5
    Private Const PhaseCount As Integer = Phase.Positive180 + 1

    Private mix As Single = 0.75F
    Private delay As Single = 0.5F
    Private delayLength As Integer
    Private depth As Single = 0.0003F
    Private feedbackLevel As Single = 0.05F
    Private frequency As Single = 1
    Private m_phase As Phase = Phase.Positive90
    Private leftDelayLine As Single() = New Single(DelayLineLength - 1) {}
    Private rightDelayLine As Single() = New Single(DelayLineLength - 1) {}
    Private writeIndex As Integer
    Private sine As Single = 0
    Private cosine As Single = 1
    Private f As Single  ' Sine/Cosine coefficient.

    Public Sub New(ByVal sampleRate As SampleRate, ByVal inputBuffer As StereoBuffer)
        MyBase.New(sampleRate, inputBuffer)
        If sampleRate Is Nothing Then
            Throw New ArgumentNullException("sampleRate")
        End If
        AddHandler sampleRate.SampleRateChanged, AddressOf HandleSampleRateChanged
        InitializeLfo()
        InitializeDelay()
    End Sub

    Public Sub New(ByVal sampleRate As SampleRate, ByVal inputBuffer As StereoBuffer, ByVal name As String)
        MyBase.New(sampleRate, inputBuffer, name)
        If sampleRate Is Nothing Then
            Throw New ArgumentNullException("sampleRate")
        End If
        AddHandler sampleRate.SampleRateChanged, AddressOf HandleSampleRateChanged
        InitializeLfo()
        InitializeDelay()
    End Sub


    Private Sub InitializeLfo()
        Dim hz As Single = frequency * FrequencyScaler
        f = PI2 * (hz / SamplesPerSecond)
    End Sub

    Private Sub InitializeDelay()
        delayLength = CInt(Math.Truncate(delay * (DelayLineLength \ 4)))
        writeIndex = 0
    End Sub

    Private Sub HandleSampleRateChanged(ByVal sender As Object, ByVal e As EventArgs)
        InitializeLfo()
    End Sub

    Public Overloads Overrides Sub Process()
        If mix < 0.00001 Then
            Return
        End If

        Dim input As Single(,) = Buffer.GetBuffer()
        Dim length As Integer = Buffer.BufferSize
        Dim readIndex As Single
        Dim delayOutput As Single

        For i As Integer = 0 To length - 1
            IncrementLfo()

            readIndex = CalculateReadIndex(GetLfoOutputLeft())

            delayOutput = GetDelayOutput(readIndex, leftDelayLine)
            leftDelayLine(writeIndex) = input(0, i) + delayOutput * feedbackLevel
            input(0, i) += delayOutput * mix

            readIndex = CalculateReadIndex(GetLfoOutputRight())

            delayOutput = GetDelayOutput(readIndex, rightDelayLine)
            rightDelayLine(writeIndex) = input(1, i) + delayOutput * feedbackLevel
            input(1, i) += delayOutput * mix

            writeIndex = (writeIndex + 1) And DelayLineMask
        Next
    End Sub

    Public Overloads Overrides Sub Reset()
        Array.Clear(leftDelayLine, 0, leftDelayLine.Length)
        Array.Clear(rightDelayLine, 0, rightDelayLine.Length)
    End Sub

    Private Sub IncrementLfo()
        sine = sine + f * cosine
        cosine = cosine - f * sine
    End Sub

    Private Function CalculateReadIndex(ByVal lfoOutput As Single) As Single
        Dim offset As Single = PowerOfTwoTable.GetPower(lfoOutput * depth) * delayLength
        Dim readIndex As Single = writeIndex - offset
        If readIndex < 0 Then
            readIndex += DelayLineLength
        End If
        If readIndex >= DelayLineLength Then readIndex = DelayLineLength - 1
        Return readIndex
    End Function

    Private Function GetLfoOutputLeft() As Single
        Dim lfoOutput As Single = 0
        Select Case m_phase
            Case Phase.Zero, Phase.Positive90, Phase.Positive180 : lfoOutput = sine
            Case Phase.Negative180 : lfoOutput = -sine
            Case Phase.Negative90 : lfoOutput = -cosine
            Case Else : Debug.Fail("Unhandled phase setting")
        End Select
        Return lfoOutput
    End Function

    Private Function GetLfoOutputRight() As Single
        Dim lfoOutput As Single = 0
        Select Case m_phase
            Case Phase.Zero : lfoOutput = sine
            Case Phase.Positive90 : lfoOutput = cosine
            Case Phase.Positive180 : lfoOutput = -sine
            Case Phase.Negative180, Phase.Negative90 : lfoOutput = sine
            Case Else : Debug.Fail("Unhandled phase setting")
        End Select
        Return lfoOutput
    End Function

    Private Shared Function GetDelayOutput(ByVal readIndex As Single, ByVal delayLine As Single()) As Single
        Dim n As Integer = CInt(Math.Truncate(readIndex))
        Dim x1 As Single = delayLine(n)
        Dim x2 As Single = delayLine((n + 1) And DelayLineMask)
        Return Synthesizer.LinearInterpolate(readIndex, x1, x2)
    End Function


#Region "IProgrammable Members"

    Public Function GetParameterName(ByVal index As Integer) As String Implements IProgrammable.GetParameterName
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Return "Chorus " & CType(index, ParameterId).ToString()
    End Function

    Public Function GetParameterLabel(ByVal index As Integer) As String Implements IProgrammable.GetParameterLabel
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Dim result As String = ""
        Select Case CType(index, ParameterId)
            Case ParameterId.Feedback : result = "Depth"
            Case ParameterId.Delay : result = "Milliseconds"
            Case ParameterId.Depth : result = "Level"
            Case ParameterId.Frequency : result = "Hz"
            Case ParameterId.Phase : result = "Angle"
            Case ParameterId.Waveform ': result = "Waveform"
            Case ParameterId.WetDryMix : result = "Level"
            Case Else : Debug.Fail("Unhandled parameter")
        End Select
        Return result
    End Function

    Public Function GetParameterDisplay(ByVal index As Integer) As String Implements IProgrammable.GetParameterDisplay
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Dim result As String = ""
        Select Case CType(index, ParameterId)

            Case ParameterId.Feedback
                result = Synthesizer.DecibelToString(feedbackLevel)

            Case ParameterId.Delay
                If True Then
                    Dim milliseconds As Single = CSng(delayLength) / SamplesPerSecond
                    milliseconds *= 1000
                    result = milliseconds.ToString(Synthesizer.FormatInfo)
                End If

            Case ParameterId.Frequency
                If True Then
                    Dim hz As Single = frequency * FrequencyScaler
                    result = hz.ToString(Synthesizer.FormatInfo)
                End If

            Case ParameterId.Depth
                result = Synthesizer.DecibelToString(depth)

            Case ParameterId.Phase
                result = m_phase.ToString()

            Case ParameterId.Waveform

            Case ParameterId.WetDryMix
                result = Synthesizer.DecibelToString(mix)

            Case Else
                Debug.Fail("Unhandled parameter")

        End Select
        Return result
    End Function

    Public Function GetParameterValue(ByVal index As Integer) As Single Implements IProgrammable.GetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Dim result As Single = 0
        Select Case CType(index, ParameterId)
            Case ParameterId.Feedback : result = feedbackLevel
            Case ParameterId.Delay : result = delay
            Case ParameterId.Depth : result = depth / DepthScaler
            Case ParameterId.Frequency : result = frequency
            Case ParameterId.Phase : result = CSng(m_phase / (PhaseCount - 1))
            Case ParameterId.Waveform
            Case ParameterId.WetDryMix : result = mix
            Case Else : Debug.Fail("Unhandled parameter")
        End Select
        Return result
    End Function

    Public Sub SetParameterValue(ByVal index As Integer, ByVal value As Single) Implements IProgrammable.SetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        ElseIf value < 0 OrElse value > 1 Then
            Throw New ArgumentOutOfRangeException("value")
        End If
        Select Case CType(index, ParameterId)
            Case ParameterId.Feedback : feedbackLevel = value
            Case ParameterId.Frequency
                frequency = value
                InitializeLfo()

            Case ParameterId.Depth : depth = value * DepthScaler

            Case ParameterId.Delay
                delay = value
                InitializeDelay()

            Case ParameterId.Phase : m_phase = CType(Math.Round((PhaseCount - 1) * value), Phase)

            Case ParameterId.Waveform

            Case ParameterId.WetDryMix : mix = value

            Case Else : Debug.Fail("Unhandled parameter")
        End Select
    End Sub

    Public ReadOnly Property ParameterCount() As Integer Implements IProgrammable.ParameterCount
        Get
            Return ParameterId.WetDryMix + 1
        End Get
    End Property

#End Region

End Class
