

'Imports System.Collections.Generic
Imports System.Diagnostics

Public Class Echo
	Inherits EffectComponent
    Implements IProgrammable

    Public Enum ParameterId
        Feedback
        LeftDelay
        PanDelay
        RightDelay
        WetDryMix
    End Enum

    ' ------------------------------------------------------------------- masks the index so that it wraps around 
    '                                                                      the delay lines creating a circular buffer
    Private Const DelayLineMask As Integer = 65535
    ' ------------------------------------------------------------------- the length of both left and right delay lines
    Private Const DelayLineLength As Integer = DelayLineMask + 1

    ' ------------------------------------------------------------------- how much to mix the effect with the dry signal
    Private mix As Single = 0.75F
    ' ------------------------------------------------------------------- Controls feedback.
    Private feedbackLevel As Single = 0.75F
    ' ------------------------------------------------------------------- left delay time
    Private leftDelay As Single = 0.5F
    ' ------------------------------------------------------------------- right delay time
    Private rightDelay As Single = 0.25F
    ' ------------------------------------------------------------------- indicates whether the delay pans 
    '                                                                     back and forth across the stereo spectrum.
    Private panDelay As Boolean = True
    ' ------------------------------------------------------------------- the left delay line
    Private leftDelayLine As Single() = New Single(DelayLineLength - 1) {}
    ' ------------------------------------------------------------------- the right delay line
    Private rightDelayLine As Single() = New Single(DelayLineLength - 1) {}
    ' ------------------------------------------------------------------- How much to offset into the left delay line 
    '                                                                     based on the left delay time
    Private leftDelayLineOffset As Integer
    ' ------------------------------------------------------------------- How much to offset into the right delay line
    '                                                                     based on the right delay time
    Private rightDelayLineOffset As Integer
    ' ------------------------------------------------------------------- current index for left and right delay lines
    Private index As Integer


    Public Sub New(ByVal sampleRate As SampleRate, ByVal buffer As StereoBuffer)
        MyBase.New(sampleRate, buffer)
        Initialize()
    End Sub

    Public Sub New(ByVal sampleRate As SampleRate, ByVal buffer As StereoBuffer, ByVal name As String)
        MyBase.New(sampleRate, buffer, name)
        Initialize()
    End Sub

    Private Sub Initialize()
        leftDelayLineOffset = CInt(Math.Truncate(DelayLineLength * (leftDelay * leftDelay)))
        rightDelayLineOffset = CInt(Math.Truncate(DelayLineLength * (rightDelay * rightDelay)))
        index = 0
    End Sub

    ' ---------------------------------------------------------------------- process the EchoEffect
    Public Overloads Overrides Sub Process()
        If mix < 0.0001 Then
            Return
        End If

        Dim input As Single(,) = Buffer.GetBuffer()

        ' ------------------------------------------------------------------ offset index into left delay line
        Dim li As Integer = (index + leftDelayLineOffset) And DelayLineMask

        ' ------------------------------------------------------------------ offset index into right delay line
        Dim ri As Integer = (index + rightDelayLineOffset) And DelayLineMask

        ' ------------------------------------------------------------------ Size of the buffer.
        Dim bufferSize As Integer = Buffer.BufferSize

        ' ------------------------------------------------------------------ Used for feeding delay back into itself.
        Dim feedback As Single

        '
        ' The algorithms for using pan delay is slightly different than 
        ' the algorithm that does not use pan delay. Pan delay simply
        ' takes the output of one delay line and feeds it into the 
        ' other delay line. The two versions of the delay algorithm
        ' are put into seperate loops for efficiency reasons.
        '

        If panDelay Then
            ' ------------------------------------------------------------------ if we're using pan delay...
            For i As Integer = 0 To bufferSize - 1
                feedback = rightDelayLine(index) * feedbackLevel
                leftDelayLine(li) = input(0, i) * mix + feedback

                feedback = leftDelayLine(index) * feedbackLevel
                rightDelayLine(ri) = input(1, i) * mix + feedback

                input(0, i) += leftDelayLine(index)
                input(1, i) += rightDelayLine(index)

                li = (li + 1) And DelayLineMask
                ri = (ri + 1) And DelayLineMask
                index = (index + 1) And DelayLineMask
            Next
        Else
            ' ------------------------------------------------------------------- else we're not using pan delay...
            For i As Integer = 0 To bufferSize - 1
                feedback = leftDelayLine(index) * feedbackLevel
                leftDelayLine(li) = input(0, i) * mix + feedback

                feedback = rightDelayLine(index) * feedbackLevel
                rightDelayLine(ri) = input(1, i) * mix + feedback

                input(0, i) += leftDelayLine(index)
                input(1, i) += rightDelayLine(index)

                li = (li + 1) And DelayLineMask
                ri = (ri + 1) And DelayLineMask
                index = (index + 1) And DelayLineMask
            Next
        End If
    End Sub


    ' ------------------------------------------------------------------------------ resets the EchoEffect
    Public Overloads Overrides Sub Reset()
        Array.Clear(leftDelayLine, 0, leftDelayLine.Length)
        Array.Clear(rightDelayLine, 0, rightDelayLine.Length)
    End Sub



#Region "IProgrammable Members"

    Public Function GetParameterName(ByVal index As Integer) As String Implements IProgrammable.GetParameterName
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Return "Echo " & CType(index, ParameterId).ToString()
    End Function

    Public Function GetParameterLabel(ByVal index As Integer) As String Implements IProgrammable.GetParameterLabel
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Dim result As String = ""
        Select Case CType(index, ParameterId)
            Case ParameterId.Feedback : result = "Depth"
            Case ParameterId.LeftDelay : result = "Seconds"
            Case ParameterId.PanDelay : result = "Pan"
            Case ParameterId.RightDelay : result = "Seconds"
            Case ParameterId.WetDryMix : result = "Level"
            Case Else : Debug.Fail("Unhandled parameter")
        End Select
        Return result
    End Function

    Public Function GetParameterDisplay(ByVal index As Integer) As String Implements IProgrammable.GetParameterDisplay
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Dim result As String = ""
        Select Case CType(index, ParameterId)
            Case ParameterId.Feedback
                result = Synthesizer.DecibelToString(feedbackLevel)

            Case ParameterId.LeftDelay
                If True Then
                    Dim milliseconds As Single = leftDelay * DelayLineLength / SamplesPerSecond
                    result = milliseconds.ToString(Synthesizer.FormatInfo)
                End If

            Case ParameterId.PanDelay
                result = (If(panDelay, "On", "Off"))

            Case ParameterId.RightDelay
                If True Then
                    Dim milliseconds As Single = rightDelay * DelayLineLength / SamplesPerSecond
                    result = milliseconds.ToString(Synthesizer.FormatInfo)
                End If

            Case ParameterId.WetDryMix
                result = Synthesizer.DecibelToString(mix)

            Case Else
                Debug.Fail("Unhandled parameter")

        End Select

        Return result
    End Function

    Public Function GetParameterValue(ByVal index As Integer) As Single Implements IProgrammable.GetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Dim result As Single = 0
        Select Case CType(index, ParameterId)
            Case ParameterId.Feedback : result = feedbackLevel
            Case ParameterId.LeftDelay : result = leftDelay
            Case ParameterId.PanDelay : result = (If(panDelay, 1, 0))
            Case ParameterId.RightDelay : result = rightDelay
            Case ParameterId.WetDryMix : result = mix
            Case Else : Debug.Fail("Unhandled parameter")
        End Select
        Return result
    End Function

    Public Sub SetParameterValue(ByVal index As Integer, ByVal value As Single) Implements IProgrammable.SetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        ElseIf value < 0 OrElse value > 1 Then
            Throw New ArgumentOutOfRangeException("value")
        End If

        Select Case CType(index, ParameterId)
            Case ParameterId.Feedback
                feedbackLevel = value

            Case ParameterId.LeftDelay
                leftDelay = value
                leftDelayLineOffset = CInt(Math.Truncate(DelayLineLength * (leftDelay * leftDelay)))

            Case ParameterId.PanDelay
                panDelay = (If(value < 0.5, False, True))

            Case ParameterId.RightDelay
                rightDelay = value
                rightDelayLineOffset = CInt(Math.Truncate(DelayLineLength * (rightDelay * rightDelay)))

            Case ParameterId.WetDryMix
                mix = value

            Case Else
                Debug.Fail("Unhandled parameter")
        End Select
    End Sub

    Public ReadOnly Property ParameterCount() As Integer Implements IProgrammable.ParameterCount
        Get
            Return ParameterId.WetDryMix + 1
        End Get
    End Property

#End Region

End Class
