
Imports System.Collections.Generic
Imports System.Diagnostics
Imports LC.Midi

' ---------------------------- a method for creating a Voice used by a Synthesizer
Public Delegate Function VoiceFactory(sampleRate As SampleRate, buffer As StereoBuffer, name As String) As Voice

' ------------------------------------------------------------------------ a voice within a Synthesizer
Public MustInherit Class Voice
	Inherits StereoSynthComponent
	Implements IProgrammable
	Implements IBendable
	Implements IControllable

    Private m_currentNote As Integer = 69
    Private components As New List(Of SynthComponent)()
    Private programmables As New ProgrammableCollection()
    Private bendables As New List(Of IBendable)()
    Private controllables As New List(Of IControllable)()
    Private m_pitchBendModulation As Single = 0
    Private mainBuffer As StereoBuffer

    Protected Sub New(ByVal sampleRate As SampleRate, ByVal buffer As StereoBuffer)
        MyBase.New(sampleRate, buffer)
        Me.mainBuffer = buffer

        AddHandler buffer.BufferSizeChanged, AddressOf HandleBufferSizeChanged
    End Sub

    Protected Sub New(ByVal sampleRate As SampleRate, ByVal buffer As StereoBuffer, ByVal name As String)
        MyBase.New(sampleRate, buffer, name)
        Me.mainBuffer = buffer
        AddHandler buffer.BufferSizeChanged, AddressOf HandleBufferSizeChanged
    End Sub

    Public Overloads Overrides Sub Synthesize(ByVal offset As Integer, ByVal count As Integer)
        If Not IsPlaying Then Return
        ' ---------------------------------------------------------------- synthesize the output for all components
        For Each component As SynthComponent In components
            component.Synthesize(offset, count)
        Next
    End Sub

    Public Overloads Overrides Sub Trigger(ByVal previousNote As Integer, ByVal note As Integer, ByVal velocity As Single)
        m_currentNote = note
        ' ---------------------------------------------------------------- trigger all components
        For Each component As SynthComponent In components
            component.Trigger(previousNote, note, velocity)
        Next
    End Sub

    Public Overloads Overrides Sub Release(ByVal velocity As Single)
        ' ---------------------------------------------------------------- release all components
        For Each component As SynthComponent In components
            component.Release(velocity)
        Next
    End Sub

    Private Sub HandleBufferSizeChanged(ByVal sender As Object, ByVal e As EventArgs)
        ' ---------------------------------------------------------------- update the buffer size for all components
        For Each component As SynthComponent In components
            component.BufferSize = mainBuffer.BufferSize
        Next
    End Sub

    ' -------------------------------------------------------------------- sort components by Ordinal value
    Private Function Comparer(ByVal a As SynthComponent, ByVal b As SynthComponent) As Integer
        If a.Ordinal > b.Ordinal Then
            Return 1
        ElseIf a.Ordinal < b.Ordinal Then
            Return -1
        Else
            Return 0
        End If
    End Function

    ' -------------------------------------------------------------------- add a SynthComponent to the Voice
    Protected Sub AddComponent(ByVal component As SynthComponent)
        If component Is Nothing Then
            Throw New ArgumentNullException("component")
        ElseIf components.Contains(component) Then
            Throw New ArgumentException("ISynthComponent has already been added to voice.")
        End If
        ' ---------------------------------------------------------------- if the component's Ordinal value changes
        '                                                                  thr component collection must be resorted
        AddHandler component.OrdinalChanged, _
                    AddressOf HandleOrdinalChanged
        ' ---------------------------------------------------------------- add component to component collection
        components.Add(component)
        ' ---------------------------------------------------------------- sort components to ensure that they
        '                                                                  synthesize their output in the right order
        components.Sort(AddressOf Comparer)
        ' ---------------------------------------------------------------- component buffer size as Voice buffer size
        component.BufferSize = mainBuffer.BufferSize
    End Sub

    ' -------------------------------------------------------------------- Removes a SynthComponent from the Voice
    Protected Sub RemoveComponent(ByVal component As SynthComponent)
        If component Is Nothing Then
            Throw New ArgumentNullException("component")
        End If
        ' ---------------------------------------------------------------- if the component was added to the Voice
        If components.Contains(component) Then
            ' ------------------------------------------------------------ disconnect from OrdinalChanged event
            RemoveHandler component.OrdinalChanged, _
                            AddressOf HandleOrdinalChanged
            ' ------------------------------------------------------------ remove component.
            components.Remove(component)
        End If
    End Sub

    Private Sub HandleOrdinalChanged(ByVal sender As Object, ByVal e As EventArgs)
        components.Sort(AddressOf Comparer)
    End Sub

    Protected Sub AddParameters(ByVal programmable As IProgrammable)
        If programmable Is Nothing Then
            Throw New ArgumentNullException("programmable")
        ElseIf programmables.Contains(programmable) Then
            Throw New ArgumentException("Parameters have already been added to Voice.")
        End If
        programmables.Add(programmable)
    End Sub

    Protected Sub AddBendable(ByVal bendable As IBendable)
        If bendable Is Nothing Then
            Throw New ArgumentNullException("bendable")
        ElseIf bendables.Contains(bendable) Then
            Throw New ArgumentException("IBendable has already been added to voice.")
        End If
        bendables.Add(bendable)
    End Sub

    Protected Sub AddControllable(ByVal controllable As IControllable)
        If controllable Is Nothing Then
            Throw New ArgumentNullException("controllable")
        ElseIf controllables.Contains(controllable) Then
            Throw New ArgumentException("IControllable has already been added to voice.")
        End If
        controllables.Add(controllable)
    End Sub


#Region "Properties"

    ' -------------------------------------------------------------------- gets the note the Voice is playing
    Public ReadOnly Property CurrentNote() As Integer
        Get
            Return m_currentNote
        End Get
    End Property

    Protected MustOverride ReadOnly Property IsPlaying() As Boolean

    Public Overloads Overrides ReadOnly Property Ordinal() As Integer
        Get
            Return 1
        End Get
    End Property

#End Region

#Region "IProgrammable Members"

    Public Function GetParameterName(ByVal index As Integer) As String Implements IProgrammable.GetParameterName
        Return programmables.GetParameterName(index)
    End Function

    Public Function GetParameterLabel(ByVal index As Integer) As String Implements IProgrammable.GetParameterLabel
        Return programmables.GetParameterLabel(index)
    End Function

    Public Function GetParameterDisplay(ByVal index As Integer) As String Implements IProgrammable.GetParameterDisplay
        Return programmables.GetParameterDisplay(index)
    End Function

    Public Function GetParameterValue(ByVal index As Integer) As Single Implements IProgrammable.GetParameterValue
        Return programmables.GetParameterValue(index)
    End Function

    Public Sub SetParameterValue(ByVal index As Integer, ByVal value As Single) Implements IProgrammable.SetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        programmables.SetParameterValue(index, value)
    End Sub

    Public ReadOnly Property ParameterCount() As Integer Implements IProgrammable.ParameterCount
        Get
            Return programmables.ParameterCount
        End Get
    End Property

#End Region

#Region "IBendable Members"

    Public Property PitchBendModulation() As Single Implements IBendable.PitchBendModulation
        Get
            Return m_pitchBendModulation
        End Get
        Set(ByVal value As Single)
            m_pitchBendModulation = value
            For Each bendable As IBendable In bendables
                bendable.PitchBendModulation = value
            Next
        End Set
    End Property

#End Region

#Region "IControllable Members"

    Public Sub ProcessControllerMessage(ByVal controllerType As ControllerType, ByVal value As Single) Implements IControllable.ProcessControllerMessage
        For Each Controllable As IControllable In controllables
            Controllable.ProcessControllerMessage(controllerType, value)
        Next
    End Sub

#End Region

End Class
