

Imports System.Collections.Generic
Imports System.Diagnostics
Imports LC.Midi


' --------------------------------------- dynamically allocating voices in response to MIDI note-on and note-off messages

Friend NotInheritable Class VoiceAllocator
    Implements IProgrammable

    Public Enum ParameterId
        PolyphonyMode
    End Enum

    ' -------------------------------------------------------------------- The list of all voices
    Private voices As List(Of Voice)
    ' -------------------------------------------------------------------- The list of voices that have been triggered
    Private triggeredVoices As List(Of Voice)
    ' -------------------------------------------------------------------- The list of voices that have been released
    Private releasedVoices As List(Of Voice)
    Private messageQueue As New Queue(Of ChannelMessage)()
    ' -------------------------------------------------------------------- The list of note-off messages that are being delayed because
    ' -------------------------------------------------------------------- the sustain pedal is currently depressed
    Private delayedNoteOffMessages As List(Of ChannelMessage)
    ' -------------------------------------------------------------------- Indicates whether the sustain pedal has been depressed
    Private m_sustainEnabled As Boolean = False
    ' -------------------------------------------------------------------- For tracking currently playing notes.
    Private tracker As New NoteTracker()
    ' -------------------------------------------------------------------- The type of polyphony to use.
    Private mode As PolyMode = PolyMode.Polyphonic


    Public Sub New(ByVal voices As IEnumerable(Of Voice))
        Me.voices = New List(Of Voice)(voices)
        Me.releasedVoices = New List(Of Voice)(voices)
        Me.triggeredVoices = New List(Of Voice)()
        Me.delayedNoteOffMessages = New List(Of ChannelMessage)()
    End Sub


    Public Sub Allocate(ByVal message As ChannelMessage)
        messageQueue.Enqueue(message)
        ProcessMessages()
    End Sub

    Public Sub AllSoundOff()
        messageQueue.Clear()
        tracker.Clear()
        delayedNoteOffMessages.Clear()
        For Each voice As Voice In triggeredVoices
            releasedVoices.Add(voice)
        Next
        triggeredVoices.Clear()
    End Sub

    Private Sub ProcessMessages()
        While messageQueue.Count > 0
            ProcessMessage(messageQueue.Dequeue())
        End While
    End Sub

    Private Sub ProcessMessage(ByVal message As ChannelMessage)
        ' If the message is a note-on message.
        If message.Command = ChannelCommand.NoteOn Then
            ' If the message is in fact a note-on message.
            If message.Data2 > 0 Then
                ProcessNoteOnMessage(message)
            Else
                ' Else the message is a note-off message. Note-on messages
                ' with a velocity of 0 are really note-off messages.
                ProcessNoteOffMessage(message)
            End If
            ' Else if the message is a note-off message.
        ElseIf message.Command = ChannelCommand.NoteOff Then
            ProcessNoteOffMessage(message)
        End If
    End Sub

    Private Sub ProcessNoteOnMessage(ByVal message As ChannelMessage)
        Debug.Assert(message.Command = ChannelCommand.NoteOn AndAlso message.Data2 > 0)

        Dim velocity As Single = CSng(message.Data2) / ChannelMessage.DataMaxValue

        ' ===========================================================================================
        '  If the note is playing then retrigger the voice with the new velocity (volume change)
        ' ===========================================================================================
        ' Find the index into the triggered voice list of the voice
        ' that is playing the specified note.
        Dim index As Integer = -1
        For i As Int32 = 0 To triggeredVoices.Count - 1
            If triggeredVoices(i).CurrentNote = message.Data1 Then
                index = i
                Exit For
            End If
        Next
        ' If a voice was playing the specified note.
        If index >= 0 Then
            Dim voice As Voice = triggeredVoices(index)
            voice.Trigger(tracker.LastNote, message.Data1, velocity)
            'voice.Trigger(message.Data1, message.Data1, velocity)
            Return
        End If

        ' ===========================================================================================
        '  If the note is not playing then trigger a new voice or the "oldest" note
        ' ===========================================================================================
        ' If there are any voices in the released voice list.
        If releasedVoices.Count > 0 Then
            ' Get voice that has been released the longest.
            Dim voice As Voice = releasedVoices(0)

            ' Remove voice from the released voice list.
            releasedVoices.RemoveAt(0)

            ' Trigger voice.
            voice.Trigger(tracker.LastNote, message.Data1, velocity)

            ' Since the voice has been triggered, add it to the triggered
            ' voice list.
            triggeredVoices.Add(voice)
            ' Else if there voices in the triggered voice list.
        ElseIf triggeredVoices.Count > 0 Then
            ' Get voice that has been triggered the longest (we're "stealing" a voice).
            Dim voice As Voice = triggeredVoices(0)


            ' Remove voice from triggered voice list.
            triggeredVoices.RemoveAt(0)

            ' Trigger voice.
            voice.Trigger(tracker.LastNote, message.Data1, velocity)

            ' Add voice to the front of the triggered voice list. 
            triggeredVoices.Add(voice)
        End If

        ' Pass message along to the note tracker.
        tracker.NoteOn(message)
    End Sub

    Private Sub ProcessNoteOffMessage(ByVal message As ChannelMessage)
        Debug.Assert(message.Command = ChannelCommand.NoteOff OrElse (message.Command = ChannelCommand.NoteOn AndAlso message.Data2 = 0))

        Dim velocity As Single = CSng(message.Data2) / ChannelMessage.DataMaxValue

        ' If the sustain pedal has been depressed.
        If m_sustainEnabled Then
            ' Delay processing note-off messages.
            delayedNoteOffMessages.Add(message)
        Else
            ' Else the sustain pedal has not been depressed.
            ' Pass message along to note tracker.
            tracker.NoteOff(message)

            ' Find the index into the triggered voice list of the voice
            ' that is playing the specified note.
            Dim index As Integer = -1
            For i As Int32 = 0 To triggeredVoices.Count - 1
                If triggeredVoices(i).CurrentNote = message.Data1 Then
                    index = i
                    Exit For
                End If
            Next


            ' If a voice was playing the specified note.
            If index >= 0 Then
                ' Get the voice that is playing the specified note.
                Dim voice As Voice = triggeredVoices(index)

                ' If mode is polyphonic.
                If mode = PolyMode.Polyphonic Then
                    ' Release voice.
                    voice.Release(velocity)

                    ' Remove voice from triggered voice list since it has 
                    ' been released.
                    triggeredVoices.RemoveAt(index)

                    ' Add voice to the released voice list since it has
                    ' been released.
                    releasedVoices.Add(voice)
                Else
                    ' Else mode is monophonic.
                    ' If a note played previously to the specified note
                    ' is still playing.
                    If tracker.Count > 0 Then
                        ' Instead of releasing voice, retrigger it with the 
                        ' previous playing note. This is done in order to mimic
                        ' how a monophonic analog synth behaves.
                        voice.Trigger(voice.CurrentNote, tracker.LastNote, 1)
                    Else
                        ' Else no other note is playing.
                        ' Release voice.
                        voice.Release(velocity)

                        ' Remove voice from triggered voice list since it has 
                        ' been released.
                        triggeredVoices.RemoveAt(index)

                        ' Add voice to the released voice list since it has
                        ' been released.
                        releasedVoices.Add(voice)
                    End If
                End If
            End If
        End If
    End Sub


    ' ------------------------------------------------------ Release all currently triggered voices
    Private Sub Reset()
        '
        For Each voice As Voice In triggeredVoices
            voice.Release(0)
        Next
        ' -------------------------------------------------- Clear all lists.
        delayedNoteOffMessages.Clear()
        messageQueue.Clear()
        triggeredVoices.Clear()
        releasedVoices.Clear()
        ' -------------------------------------------------- If mode is polyphonic.
        If mode = PolyMode.Polyphonic Then
            ' Place all voices in the released list. As new notes are
            ' played, voices will be drawn from this list.
            For Each voice As Voice In voices
                releasedVoices.Add(voice)
            Next
        Else
            ' ---------------------------------------------- Else mode is monophonic.
            ' If there are any voices.
            If voices.Count > 0 Then
                ' Add only a single voice to the released list. This 
                ' is the voice that will be used to play monophonically.
                releasedVoices.Add(voices(0))
            End If
        End If
    End Sub



    Public WriteOnly Property SustainEnabled() As Boolean
        Set(ByVal value As Boolean)
            If value = m_sustainEnabled Then
                Return
            End If
            m_sustainEnabled = value
            If Not m_sustainEnabled Then
                ' Place all delayed note-off messages into MIDI queue. 
                ' They can now be processed.
                For Each noteOffMessage As ChannelMessage In delayedNoteOffMessages
                    messageQueue.Enqueue(noteOffMessage)
                Next
                delayedNoteOffMessages.Clear()
                ProcessMessages()
            End If
        End Set
    End Property



#Region "IProgrammable Members"

    Public Function GetParameterName(ByVal index As Integer) As String Implements IProgrammable.GetParameterName
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Dim result As String = ""
        Select Case CType(index, ParameterId)
            Case ParameterId.PolyphonyMode
                result = "Polyphony Mode"
            Case Else
                Debug.Fail("Unhandled parameter")
        End Select
        Return result
    End Function

    Public Function GetParameterLabel(ByVal index As Integer) As String Implements IProgrammable.GetParameterLabel
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Dim result As String = ""
        Select Case CType(index, ParameterId)
            Case ParameterId.PolyphonyMode
                result = "Poly/Mono"
            Case Else
                Debug.Fail("Unhandled parameter")
        End Select
        Return result
    End Function

    Public Function GetParameterDisplay(ByVal index As Integer) As String Implements IProgrammable.GetParameterDisplay
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Dim result As String = ""
        Select Case CType(index, ParameterId)
            Case ParameterId.PolyphonyMode
                result = mode.ToString()
            Case Else
                Debug.Fail("Unhandled parameter")
        End Select
        Return result
    End Function

    Public Function GetParameterValue(ByVal index As Integer) As Single Implements IProgrammable.GetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Dim result As Single = 0
        Select Case CType(index, ParameterId)
            Case ParameterId.PolyphonyMode
                If mode = PolyMode.Polyphonic Then
                    result = 0
                Else
                    result = 1
                End If
            Case Else
                Debug.Fail("Unhandled parameter")
        End Select
        Return result
    End Function

    Public Sub SetParameterValue(ByVal index As Integer, ByVal value As Single) Implements IProgrammable.SetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        ElseIf value < 0 OrElse value > 1 Then
            Throw New ArgumentOutOfRangeException("value")
        End If
        Select Case CType(index, ParameterId)
            Case ParameterId.PolyphonyMode
                mode = (If(value < 0.5, PolyMode.Polyphonic, PolyMode.Monophonic))
                Reset()
            Case Else
                Debug.Fail("Unhandled parameter")
        End Select
    End Sub

    Public ReadOnly Property ParameterCount() As Integer Implements IProgrammable.ParameterCount
        Get
            Return ParameterId.PolyphonyMode + 1
        End Get
    End Property

#End Region

End Class
