

Imports System.ComponentModel
Imports System.Diagnostics
Imports System.Threading
Imports LC.Collections.Generic

' --------------------------------------------------------------------------- an asynchronous queue of delegates
Public Partial Class DelegateQueue
	Inherits SynchronizationContext
    Implements IComponent
    Implements ISynchronizeInvoke


    ' ---------------------------------------------------------------------- the thread for processing delegates
    Private delegateThread As Thread
    ' ---------------------------------------------------------------------- deque for holding delegates
    Private delegateDeque As New Deque(Of DelegateQueueAsyncResult)()
    ' ---------------------------------------------------------------------- object to use for locking
    Private ReadOnly lockObject As New Object()
    ' ---------------------------------------------------------------------- synchronization context in which this DelegateQueue was created
    Private context As SynchronizationContext
    ' ---------------------------------------------------------------------- indicates whether the delegate queue has been disposed
    Private _disposed As Boolean = False
    ' ---------------------------------------------------------------------- thread ID counter for all DelegateQueues
    Private Shared threadID As UInteger = 0
    ' ----------------------------------------------------------------------
    Private m_site As ISite = Nothing


    ' ------------------------------------------------------------------------ "Events"


    ' ---------------------------------------- Occurs after a method has been invoked as a result of a call to 
    '                                            the BeginInvoke or BeginInvokePriority methods.
    '
    Public Event InvokeCompleted As EventHandler(Of InvokeCompletedEventArgs)

    ' ---------------------------------------- Occurs after a method has been invoked as a result of a call to
    '                                            the Post and PostPriority methods.
    '
    Public Event PostCompleted As EventHandler(Of PostCompletedEventArgs)


    ' ------------------------------------------------------------- "Construction"
    Public Sub New()
        InitializeDelegateQueue()

        If SynchronizationContext.Current Is Nothing Then
            context = New SynchronizationContext()
        Else
            context = SynchronizationContext.Current
        End If
    End Sub


    Public Sub New(ByVal container As IContainer)
        '''
        ''' Required for Windows.Forms Class Composition Designer support
        '''
        container.Add(Me)
        InitializeDelegateQueue()
    End Sub

    Protected Overrides Sub Finalize()
        Try

            Dispose(False)
        Finally
            MyBase.Finalize()
        End Try
    End Sub


    Private Sub InitializeDelegateQueue()
        ' --------------------------------------------------------- create thread for processing delegates
        delegateThread = New Thread(AddressOf DelegateProcedure)
        ' --------------------------------------------------------- make the thread as background thread.
        'delegateThread.IsBackground = False
        ' --------------------------------------------------------- set the Priority of the thread.
        delegateThread.Priority = ThreadPriority.Highest

        SyncLock lockObject
            ' -------------------------------------------------------------------- increment to next thread ID
            threadID += 1UI
            ' -------------------------------------------------------------------- create name for thread
            delegateThread.Name = "Delegate Queue Thread: " & threadID.ToString()
            ' -------------------------------------------------------------------- start thread
            delegateThread.Start()
            '
            'Debug.WriteLine(delegateThread.Name & " Started.")

            ' -------------------------------------------------------------------- wait signal from thread running
            Monitor.Wait(lockObject)
        End SyncLock
    End Sub


    ' ----------------------------------------------------------------------------- "Methods"

    Protected Overridable Sub Dispose(ByVal disposing As Boolean)
        If disposing Then
            SyncLock lockObject
                _disposed = True

                Monitor.Pulse(lockObject)

                GC.SuppressFinalize(Me)
            End SyncLock
        End If
    End Sub


    ' ---------------------------------- Executes the delegate on the main thread that this object executes on.
    '
    Public Function BeginInvokePriority(ByVal method As [Delegate], ByVal ParamArray args As Object()) As IAsyncResult
        If _disposed Then
            Throw New ObjectDisposedException("DelegateQueue")
        ElseIf method Is Nothing Then
            Throw New ArgumentNullException()
        End If

        Dim result As DelegateQueueAsyncResult

        ' If BeginInvokePriority was called from a different thread than the one
        ' in which the DelegateQueue is running.
        If InvokeRequired Then
            result = New DelegateQueueAsyncResult(Me, method, args, False, NotificationType.BeginInvokeCompleted)

            SyncLock lockObject
                ' Put the method at the front of the queue.
                delegateDeque.PushFront(result)

                Monitor.Pulse(lockObject)
            End SyncLock
        Else
            ' Else BeginInvokePriority was called from the same thread in which the 
            ' DelegateQueue is running.
            result = New DelegateQueueAsyncResult(Me, method, args, True, NotificationType.None)

            ' The method is invoked here instead of placing it in the 
            ' queue. The reason for this is that if EndInvoke is called 
            ' from the same thread in which the DelegateQueue is running and
            ' the method has not been invoked, deadlock will occur. 
            result.Invoke()
        End If

        Return result
    End Function


    ' --------------------------------------- executes the delegate on the main thread that this object executes on
    '
    Public Function InvokePriority(ByVal method As [Delegate], ByVal ParamArray args As Object()) As Object
        If _disposed Then
            Throw New ObjectDisposedException("DelegateQueue")
        ElseIf method Is Nothing Then
            Throw New ArgumentNullException()
        End If

        Dim returnValue As Object = Nothing

        ' If InvokePriority was called from a different thread than the one
        ' in which the DelegateQueue is running.
        If InvokeRequired Then
            Dim result As New DelegateQueueAsyncResult(Me, method, args, False, NotificationType.None)

            SyncLock lockObject
                ' Put the method at the back of the queue.
                delegateDeque.PushFront(result)
                Monitor.Pulse(lockObject)
            End SyncLock

            ' Wait for the result of the method invocation.
            returnValue = EndInvoke(result)
        Else
            ' Else InvokePriority was called from the same thread in which the 
            ' DelegateQueue is running.
            ' Invoke the method here rather than placing it in the queue.
            returnValue = method.DynamicInvoke(args)
        End If

        Return returnValue
    End Function


    ' --------------------------------------- executes the delegate on the main thread that this object executes on
    '
    Public Function BeginInvoke(ByVal callback As AsyncCallback, ByVal state As Object, ByVal method As [Delegate], ByVal ParamArray args As Object()) As IAsyncResult
        If _disposed Then
            Throw New ObjectDisposedException("DelegateQueue")
        ElseIf method Is Nothing Then
            Throw New ArgumentNullException()
        End If

        Dim result As DelegateQueueAsyncResult

        If InvokeRequired Then
            result = New DelegateQueueAsyncResult(Me, callback, state, method, args, False, NotificationType.BeginInvokeCompleted)

            SyncLock lockObject
                delegateDeque.PushBack(result)
                Monitor.Pulse(lockObject)
            End SyncLock
        Else
            result = New DelegateQueueAsyncResult(Me, callback, state, method, args, False, NotificationType.None)
            result.Invoke()
        End If

        Return result
    End Function


    ' ------------------------------------------- dispatches an asynchronous message to this synchronization context
    '
    Public Sub PostPriority(ByVal d As SendOrPostCallback, ByVal state As Object)
        If _disposed Then
            Throw New ObjectDisposedException("DelegateQueue")
        ElseIf d Is Nothing Then
            Throw New ArgumentNullException()
        End If

        SyncLock lockObject
            Dim result As New DelegateQueueAsyncResult(Me, d, New Object() {state}, False, NotificationType.PostCompleted)

            ' Put the method at the front of the queue.
            delegateDeque.PushFront(result)

            Monitor.Pulse(lockObject)
        End SyncLock
    End Sub


    ' --------------------------------------------- dispatches an synchronous message to this synchronization context 
    '
    Public Sub SendPriority(ByVal d As SendOrPostCallback, ByVal state As Object)
        InvokePriority(d, state)
    End Sub


    ' ---------------------------------------------- processes and invokes delegates
    '
    Private Sub DelegateProcedure()
        SyncLock lockObject
            ' Signal the constructor that the thread is now running.
            Monitor.Pulse(lockObject)
        End SyncLock

        ' Set this DelegateQueue as the SynchronizationContext for this thread.
        SynchronizationContext.SetSynchronizationContext(Me)

        ' Placeholder for DelegateQueueAsyncResult objects.
        Dim result As DelegateQueueAsyncResult = Nothing

        ' While the DelegateQueue has not been disposed.
        While True
            ' Critical section.
            SyncLock lockObject
                ' If the DelegateQueue has been disposed, break out of loop; we're done.
                If _disposed Then
                    Exit While
                End If

                ' If there are delegates waiting to be invoked.
                If delegateDeque.Count > 0 Then
                    result = delegateDeque.PopFront()
                Else
                    ' Else there are no delegates waiting to be invoked.
                    ' Wait for next delegate.
                    Monitor.Wait(lockObject)

                    ' If the DelegateQueue has been disposed, break out of loop; we're done.
                    If _disposed Then
                        Exit While
                    End If

                    Debug.Assert(delegateDeque.Count > 0)

                    result = delegateDeque.PopFront()
                End If
            End SyncLock

            Debug.Assert(result IsNot Nothing)

            ' Invoke the delegate.
            result.Invoke()

            If result.NotificationType = NotificationType.BeginInvokeCompleted Then
                Dim e As New InvokeCompletedEventArgs(result.Method, result.GetArgs(), result.ReturnValue, result.[Error])

                OnInvokeCompleted(e)
            ElseIf result.NotificationType = NotificationType.PostCompleted Then
                Dim args As Object() = result.GetArgs()

                Debug.Assert(args.Length = 1)
                Debug.Assert(TypeOf result.Method Is SendOrPostCallback)

                Dim e As New PostCompletedEventArgs(DirectCast(result.Method, SendOrPostCallback), result.[Error], args(0))

                OnPostCompleted(e)
            Else
                Debug.Assert(result.NotificationType = NotificationType.None)
            End If
        End While

        'Debug.WriteLine(delegateThread.Name & " Finished")
    End Sub


    ' --------------------------------------------------------- raises the InvokeCompleted event
    '
    Protected Overridable Sub OnInvokeCompleted(ByVal e As InvokeCompletedEventArgs)
        'Dim handler As EventHandler(Of InvokeCompletedEventArgs) = InvokeCompleted
        '	If handler IsNot Nothing Then
        '		context.Post(Function(state As Object) Do
        '			handler(Me, e)
        '		End Function, Nothing)
        '	End If
        context.Post(Function(state As Object) InvokeCompleted_Function(e), Nothing)
    End Sub
    Private Function InvokeCompleted_Function(ByVal e As InvokeCompletedEventArgs) As Boolean
        RaiseEvent InvokeCompleted(Me, e)
    End Function


    ' --------------------------------------------------------- raises the PostCompleted event
    '
    Protected Overridable Sub OnPostCompleted(ByVal e As PostCompletedEventArgs)
        '      Dim handler As EventHandler(Of PostCompletedEventArgs) = PostCompleted
        '      If handler IsNot Nothing Then
        '	context.Post(Function(state As Object) Do
        '          handler(Me, e)
        '	End Function, Nothing)
        'End If
        context.Post(Function(state As Object) PostCompleted_Function(e), Nothing)
    End Sub
    Private Function PostCompleted_Function(ByVal e As PostCompletedEventArgs) As Boolean
        RaiseEvent PostCompleted(Me, e)
    End Function


    ' --------------------------------------------------------- raises the Disposed event
    '
    Protected Overridable Sub OnDisposed(ByVal e As EventArgs)
        '      Dim handler As EventHandler = Disposed
        '      If handler IsNot Nothing Then
        '	context.Post(Function(state As Object) Do
        '          handler(Me, e)
        '	End Function, Nothing)
        'End If
        context.Post(Function(state As Object) OnDisposed_Function(e), Nothing)
    End Sub
    Private Function OnDisposed_Function(ByVal e As EventArgs) As Boolean
        RaiseEvent Disposed(Me, e)
    End Function


#Region "SynchronizationContext Overrides"

    ' ------------------------------------------ dispatches a synchronous message to this synchronization context 
    '
    Public Overloads Overrides Sub Send(ByVal d As SendOrPostCallback, ByVal state As Object)
        Invoke(d, CType(state, Object()))
    End Sub


    ' ------------------------------------------ dispatches an asynchronous message to this synchronization context
    '
    Public Overloads Overrides Sub Post(ByVal d As SendOrPostCallback, ByVal state As Object)
        If _disposed Then
            Throw New ObjectDisposedException("DelegateQueue")
        ElseIf d Is Nothing Then
            Throw New ArgumentNullException()
        End If

        SyncLock lockObject
            delegateDeque.PushBack(New DelegateQueueAsyncResult(Me, d, New Object() {state}, False, NotificationType.PostCompleted))

            Monitor.Pulse(lockObject)
        End SyncLock
    End Sub

#End Region

#Region "IComponent Members"

    ' ----------------------------------------------------------- handles the Disposed delegate of a DelegateQueue
    '
    Public Event Disposed(ByVal sender As Object, ByVal e As System.EventArgs) Implements System.ComponentModel.IComponent.Disposed

    ' ----------------------------------------------------------- gets or set the ISite associated with DelegateQueue
    Public Property Site() As ISite Implements IComponent.Site
        Get
            Return m_site
        End Get
        Set(ByVal value As ISite)
            m_site = value
        End Set
    End Property

#End Region

#Region "ISynchronizeInvoke Members"

    ' ----------------------------- Executes the delegate on the main thread that this DelegateQueue executes on.
    ' 
    Public Function BeginInvoke(ByVal method As System.Delegate, ByVal args() As Object) As IAsyncResult Implements ISynchronizeInvoke.BeginInvoke
        If _disposed Then
            Throw New ObjectDisposedException("DelegateQueue")
        ElseIf method Is Nothing Then
            Throw New ArgumentNullException()
        End If

        Dim result As DelegateQueueAsyncResult

        If InvokeRequired Then
            result = New DelegateQueueAsyncResult(Me, method, args, False, NotificationType.BeginInvokeCompleted)

            SyncLock lockObject
                delegateDeque.PushBack(result)

                Monitor.Pulse(lockObject)
            End SyncLock
        Else
            result = New DelegateQueueAsyncResult(Me, method, args, False, NotificationType.None)

            result.Invoke()
        End If

        Return result
    End Function


    ' ------------------------- Waits until the process started by calling BeginInvoke completes, and then returns 
    '
    Public Function EndInvoke(ByVal result As IAsyncResult) As Object Implements ISynchronizeInvoke.EndInvoke
        If _disposed Then
            Throw New ObjectDisposedException("DelegateQueue")
        ElseIf Not (TypeOf result Is DelegateQueueAsyncResult) Then
            Throw New ArgumentException()
        ElseIf DirectCast(result, DelegateQueueAsyncResult).Owner IsNot Me Then
            Throw New ArgumentException()
        End If

        result.AsyncWaitHandle.WaitOne()

        Dim r As DelegateQueueAsyncResult = DirectCast(result, DelegateQueueAsyncResult)

        If r.[Error] IsNot Nothing Then
            Throw r.[Error]
        End If

        Return r.ReturnValue
    End Function


    ' ------------------------------ executes the delegate on the main thread that this DelegateQueue executes on
    '
    Public Function Invoke(ByVal method As System.Delegate, ByVal args() As Object) As Object Implements ISynchronizeInvoke.Invoke
        If _disposed Then
            Throw New ObjectDisposedException("DelegateQueue")
        ElseIf method Is Nothing Then
            Throw New ArgumentNullException()
        End If

        Dim returnValue As Object = Nothing

        If InvokeRequired Then
            Dim result As New DelegateQueueAsyncResult(Me, method, args, False, NotificationType.None)

            SyncLock lockObject
                delegateDeque.PushBack(result)

                Monitor.Pulse(lockObject)
            End SyncLock

            returnValue = EndInvoke(result)
        Else
            ' Invoke the method here rather than placing it in the queue.
            returnValue = method.DynamicInvoke(args)
        End If

        Return returnValue
    End Function



    ' -------------------------------------------- gets a value indicating whether the caller must call Invoke
    '
    Public ReadOnly Property InvokeRequired() As Boolean Implements ISynchronizeInvoke.InvokeRequired
        Get
            Return Thread.CurrentThread.ManagedThreadId <> delegateThread.ManagedThreadId
        End Get
    End Property

#End Region

#Region "IDisposable Members"

    ' ------------------------------------------------------------- Disposes of the DelegateQueue.
    Public Sub Dispose() Implements IDisposable.Dispose
        If _disposed Then
            Return
        End If
        Dispose(True)
        OnDisposed(EventArgs.Empty)
    End Sub

#End Region


End Class
